--------------------------------------------------------------------------------
-- copyright notice:
--------------------------------------------------------------------------------
-- copyright (c) 2010 by
--
-- CESYS GmbH / GERMANY
--
-- http://www.cesys.com
--------------------------------------------------------------------------------
-- author: Manfred Kraus <Manfred.Kraus@cesys.com>
--------------------------------------------------------------------------------
--
--------------------------------------------------------------------------------
-- license:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS NOT FREE! IT IS FOR USE TOGETHER WITH CESYS PRODUCTS ONLY!
-- YOU ARE NOT ALLOWED TO MODIFY AND DISTRIBUTE OR USE IT WITH ANY OTHER
-- HARDWARE, SOFTWARE OR ANY OTHER KIND OF ASIC OR PROGRAMMABLE LOGIC DESIGN
-- WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDER!
--------------------------------------------------------------------------------
-- disclaimer of warranty:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
-- THERE IS NO WARRANTY OR SUPPORT FOR THIS SOURCECODE. THE COPYRIGHT HOLDER 
-- PROVIDES THIS SOURCECODE "AS IS" WITHOUT WARRANTY OF ANY KIND,
-- EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
-- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-- THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THIS SOURCECODE IS WITH
-- YOU. SHOULD THIS SOURCECODE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
-- NECESSARY SERVICING, REPAIR OR CORRECTION.
--
-- IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES, INCLUDING
-- ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
-- USE OR INABILITY TO USE THIS SOURCECODE (INCLUDING BUT NOT LIMITED TO LOSS OF
-- DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD 
-- PARTIES OR A FAILURE OF THIS SOURCECODE TO OPERATE WITH ANY OTHER
-- SOFTWARE-PROGRAMS, HARDWARE-CIRCUITS OR ANY OTHER KIND OF ASIC OR
-- PROGRAMMABLE LOGIC DESIGN), EVEN IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF 
-- THE POSSIBILITY OF SUCH DAMAGES.
--------------------------------------------------------------------------------

library ieee;
	use ieee.std_logic_1164.all;
	use ieee.numeric_std.all;
	use ieee.math_real.all;
    
library unisim;
use unisim.vcomponents.all;

use work.wishbone_pkg.all;
use work.wb_ma_fx3_pkg.all;
use work.wb_sl_bram_pkg.all;


entity efm02_bram_top is
  port (
    pin_fx3_flag_i   : in    std_logic_vector(3 downto 0);
    pin_fx3_slcs_n_o   : out   std_logic;
    pin_fx3_sloe_n_o   : out   std_logic;
    pin_fx3_slrd_n_o   : out   std_logic;
    pin_fx3_slwr_n_o   : out   std_logic;
    pin_fx3_pktend_n_o : out   std_logic;
    pin_fx3_fifoadr_o  : out   std_logic_vector(1 downto 0);
    pin_fx3_fd_io      : inout std_logic_vector(31 downto 0);

	pin_clk100_i	: in std_logic; -- external oscillator (default: 100 MHz)
	pin_gpifclk_o	: out std_logic; -- source of FX-3 <--> FPGA clock network on PCB
	pin_gpifclk_i	: in std_logic; -- feedback of FX-3 <--> FPGA clock network on PCB
    pin_user_led_o	: out std_logic;
	pin_sysreset_i_n : in std_logic;
    pin_debug_o: out   std_logic_vector(15 downto 0)

--    sys_rst_n : in std_logic;
--    sys_clk   : in std_logic
 
 );
end efm02_bram_top;


architecture rtl of efm02_bram_top is
  
	signal syscon : rec_syscon_port := SYSCON_DEFAULT;
	signal master : rec_master_port := MASTER_DEFAULT;
	signal slave  : rec_slave_port  := SLAVE_DEFAULT;

	signal clk100_i				: std_logic;
	signal ledclk				: std_logic;
  
	signal clk_gpif_n				: std_logic;
	signal clk_gpif					: std_logic;
	signal sys_clk  				: std_logic := '0';

	signal c3_sys_clk	 			:std_logic := '0';

	signal sys_clk_locked	: std_logic := '0';
	
	signal gpifclk_o: std_logic := '0';
	signal gpifclk_tristate: std_logic := '0';
	
	signal Freshly_configured_n : std_logic := '0';
    signal FPGA_Reset: std_logic := '1';
    signal sysreset_i : std_logic := '1';

  	constant clk_frequency_Hz			: real := 100.0e+6; -- 100 MHz
	constant BlinkingLedFrequency		: real := 4.0;  -- 2 Hz 
	constant BlinkingLedToggleCycles 	: positive := integer(ceil(clk_frequency_Hz / (2.0 * BlinkingLedFrequency)));

	signal BlinkingLed				: std_logic := '0';    
    
component clock_alignment
port
 (-- Clock in ports
  CLK_IN1           : in     std_logic;
  -- Clock out ports
  CLK_OUT1          : out    std_logic;
  -- Status and control signals
  RESET             : in     std_logic;
  LOCKED            : out    std_logic
 );
end component;

------------------------------------------------------------------------------
-- "Output    Output      Phase     Duty      Pk-to-Pk        Phase"
-- "Clock    Freq (MHz) (degrees) Cycle (%) Jitter (ps)  Error (ps)"
------------------------------------------------------------------------------
-- CLK_OUT1____50.000______0.000______50.0______219.229____166.228
-- CLK_OUT2____75.000______0.000______50.0______201.586____166.228
-- CLK_OUT3____80.769______0.000______50.0______198.610____166.228
-- CLK_OUT4____87.500______0.000______50.0______195.472____166.228
-- CLK_OUT5____95.455______0.000______50.0______192.145____166.228
--
------------------------------------------------------------------------------
-- "Input Clock   Freq (MHz)    Input Jitter (UI)"
------------------------------------------------------------------------------
-- __primary_________100.000____________0.010

component gpif_clock
port
 (-- Clock in ports
  CLK_IN1           : in     std_logic;
  -- Clock out ports
  CLK_OUT1          : out    std_logic;
  CLK_OUT2          : out    std_logic;
  CLK_OUT3          : out    std_logic;
  CLK_OUT4          : out    std_logic;
  CLK_OUT5          : out    std_logic
 );
end component;


begin  -- rtl

	syscon.clk <= sys_clk;
--    pin_user_led_o <= '1';
    
    -----------------------------------------------------------------------------------------------
	-- After configuration, stay in Reset until FX-3 pulsed Reset low.
	-----------------------------------------------------------------------------------------------   
    process (c3_sys_clk, pin_sysreset_i_n)
    begin
		if rising_edge (c3_sys_clk) then -- use a clock that is available while reset is active.
            sysreset_i <= not pin_sysreset_i_n; -- sync external Reset signal
        end if;
    end process;
    
    process (c3_sys_clk, pin_sysreset_i_n)
    begin
		if rising_edge (c3_sys_clk) then -- use a clock that is available while reset is active.
            if (pin_sysreset_i_n = '0') then
                Freshly_configured_n <= '1';
            end if;    
        end if;
    end process;

    FPGA_Reset <= sysreset_i or (not Freshly_configured_n);

    
    
    
    
---------------------------------------------------
--  Clock management and GPIF clock generation
---------------------------------------------------

-- IBUF the incoming clock (tip: this ibuffed clock can be used as clock for mcb)
  IBUFG_inst_clk100 : IBUFG
  generic map (
	 IBUF_LOW_PWR => TRUE, -- Low power (TRUE) vs. performance (FALSE) setting for referenced I/O standards
	 IOSTANDARD => "DEFAULT")
  port map (
	 O => clk100_i, -- Clock buffer output
	 I => pin_clk100_i	-- Clock buffer input (connect directly to top-level port)
  );

      
gpif_clock_generator : gpif_clock
  port map
   (-- Clock in ports
    CLK_IN1 => clk100_i,
    -- Clock out ports
    CLK_OUT1 => open,       -- 50.000 MHz
    CLK_OUT2 => open,       -- 75.000 MHz
    CLK_OUT3 => open,       -- 80.769 MHz
    CLK_OUT4 => c3_sys_clk, -- 87.500 MHz
    CLK_OUT5 => open);      -- 95.455 MHz
    
-- produce a 180° version of the clock	
  	clk_gpif <= c3_sys_clk;
	clk_gpif_n <= not c3_sys_clk;

-- use the clock and its 180° version with a ODDR2 to output the clock on a regular I/O pin.	
  inst_ODDR2_clk : ODDR2
  generic map(
	DDR_ALIGNMENT => "C0",			-- Sets output alignment to "NONE", "C0", "C1" : Input data D0, D1 is alligned to C0 = i_clk
	INIT		  => '0',			-- Sets initial state of the Q output to '0' or '1'
	SRTYPE		  => "ASYNC")		-- Specifies "SYNC" or "ASYNC" set/reset
  port map (						
	Q			  => gpifclk_o,	-- 1-bit output data
	C0			  => clk_gpif,		-- 1-bit clock input
	C1			  => clk_gpif_n,	-- 1-bit clock input
	CE			  => '1',			-- 1-bit clock enable input
	D0			  => '1',			-- 1-bit data input (associated with C0)
	D1			  => '0',			-- 1-bit data input (associated with C1)
	R			  => '0',			-- 1-bit reset input
	S			  => '0'			-- 1-bit set input
  );

pin_gpifclk_o <= gpifclk_o;

  
-------------------------------------------------------
-- produce sys_clock to drive the user fpga-fabric  
-------------------------------------------------------   
    inst_clock_alignment : clock_alignment
	 port map
   (-- Clock in ports
	CLK_IN1 => pin_gpifclk_i,
	CLK_OUT1 => sys_clk,
	RESET	=> '0',
	LOCKED => sys_clk_locked
    );    
   
   
  process(syscon)
  begin
    if syscon.clk'event and syscon.clk = '1' then
      syscon.rst <= FPGA_Reset or (not sys_clk_locked);
    end if;
  end process;

  
  	-----------------------------------------------------------------------------------------------
	-- Blink the Led !
	-----------------------------------------------------------------------------------------------

    ledclk <= c3_sys_clk;
    
	process (ledclk)
		variable cnt: INTEGER RANGE 0 to BlinkingLedToggleCycles;
	begin
		if rising_edge (ledclk) then
            -- if (syscon.rst = '1') then
                -- BlinkingLed <= '0';
                -- cnt := 0;
            -- else    
                if (cnt = 0) then
                    cnt := BlinkingLedToggleCycles ;
                    BlinkingLed <= not BlinkingLed;
                else
                    cnt := cnt - 1;
                end if;
--            end if;
		end if;
	end process;

	pin_user_led_o <= BlinkingLed;

  
  
  
  inst_wb_ma_fx3 : wb_ma_fx3
    generic map (
      NR_OF_DBGPORTS => 16
    )
    port map (
      syscon_i => syscon,
      master_o => master,
      slave_i  => slave,

      fx3_flag_i   => pin_fx3_flag_i,
      fx3_slcs_n_o   => pin_fx3_slcs_n_o,
      fx3_sloe_n_o   => pin_fx3_sloe_n_o,
      fx3_slrd_n_o   => pin_fx3_slrd_n_o,
      fx3_slwr_n_o   => pin_fx3_slwr_n_o,
      fx3_pktend_n_o => pin_fx3_pktend_n_o,
      fx3_fifoadr_o  => pin_fx3_fifoadr_o,
      fx3_fd_io      => pin_fx3_fd_io,

--      debug => open
      debug => pin_debug_o
    );

  inst_wb_sl_bram : wb_sl_bram
    generic map (
      NR_OF_DBGPORTS => 1
    )
    port map (
      syscon_i => syscon,
      master_i => master,
      slave_o  => slave,

      debug => open
    );

end rtl;
