----------------------------------------------------------------------------------
-- Description:    	This is a VHDL top-level file for the CESYS EFM-02 module
-- Design Name:		efm02_perf_top_<DEVICETYPE> (performance demo)
--
-- Company: 		CESYS GmbH
-- Author: 			Manfred Kraus <manfred.kraus@cesys.com>
-- 
-- Module Name:    	efm02_toplevel - RTL 
-- Target Devices: 	XC6SLX45-3FGG484, XC6SLX100-3FGG484, XC6SLX45-3FGG484, 
-- Tool versions:  	ISE Project Navigator 14.7
---------------------------------------------------------------------------------------------------
-- Copyright © 2013-2014 by CESYS GmbH, Germany. All rights are reserved. 
-- Unauthorized duplication of this document, in whole or in part, by any means is prohibited
-- without the prior written permission of CESYS GmbH, Germany.
-- 
-- Although CESYS GmbH believes that the information included in this publication is correct
-- as of the date of publication, it is provided "as is" without any support or guarantee.
-- CESYS reserves the right to make changes at any time without notice.
-- 
-- All information in this document may only be published by CESYS GmbH, Germany.
---------------------------------------------------------------------------------------------------
--    
---------------------------------------------------------------------------------------------------
-- File history:
--
-- Version | Date       | Author           | Remarks
-- ------------------------------------------------------------------------------------------------
-- 1.0     | 21.01.2014 | M. Kraus         | First release
---------------------------------------------------------------------------------------------------

library ieee;
    use ieee.std_logic_1164.all;
	use ieee.numeric_std.all;
	use ieee.math_real.all;
    
library unisim;
use unisim.vcomponents.all;

use work.efm02_soc_pkg.all;
use work.wishbone_pkg.all;
use work.wb_ma_fx3_pkg.all;
use work.wb_sl_bram_pkg.all;


entity efm02_perf_top_XC6SLX45 is
  port (
	pin_fx3_flag_i   : in	   std_logic_vector(3 downto 0);
	pin_fx3_slcs_n_o   : out   std_logic;
	pin_fx3_sloe_n_o   : out   std_logic;
	pin_fx3_slrd_n_o   : out   std_logic;
	pin_fx3_slwr_n_o   : out   std_logic;
	pin_fx3_pktend_n_o : out   std_logic;
	pin_fx3_fifoadr_o  : out   std_logic_vector(1 downto 0);
	pin_fx3_fd_io	   : inout std_logic_vector(31 downto 0);

--    pin_debug_o: out   std_logic_vector(15 downto 0);
    
	pin_user_led_o : out std_logic;
	
	pin_clk100_i	: in std_logic; -- external oscillator (default: 100 MHz)
	pin_gpifclk_o	: out std_logic; -- source to FX-3 <--> FPGA clock network on PCB
	pin_gpifclk_i	: in std_logic; -- feedback from FX-3 <--> FPGA clock network on PCB

	pin_sysreset_i : in std_logic
  );
end efm02_perf_top_XC6SLX45;


architecture rtl of efm02_perf_top_XC6SLX45 is
  
  signal syscon	  : rec_syscon_port		:= SYSCON_DEFAULT;
  signal intercon : rec_intercon_signal := INTERCON_DEFAULT;

	signal clk100_i				: std_logic;
	signal ledclk				: std_logic;
  
	signal clk_gpif_n			: std_logic;
	signal clk_gpif				: std_logic;
	signal sys_clk  			: std_logic := '0';
    signal fx3_plclk  			: std_logic := '0';

	signal sys_clk_locked	    : std_logic := '0';
	
	signal gpifclk_o: std_logic := '0';
	signal gpifclk_tristate: std_logic := '0';
	
    signal sysreset_i : std_logic := '1';
    

  	constant clk_frequency_Hz			: real := 100.0e+6; -- 100 MHz
	constant BlinkingLedFrequency		: real := 4.0;  -- 2 Hz 
	constant BlinkingLedToggleCycles 	: positive := integer(ceil(clk_frequency_Hz / (2.0 * BlinkingLedFrequency)));

	signal BlinkingLed				: std_logic := '0';   
  


component clock_alignment
port
 (-- Clock in ports
  CLK_IN1           : in     std_logic;
  -- Clock out ports
  CLK_OUT1          : out    std_logic;
  -- Status and control signals
  RESET             : in     std_logic;
  LOCKED            : out    std_logic
 );
end component;


component gpif_clock
port
 (-- Clock in ports
  CLK_IN1           : in     std_logic;
  -- Clock out ports
  CLK_OUT1          : out    std_logic;
  CLK_OUT2          : out    std_logic;
  CLK_OUT3          : out    std_logic;
  CLK_OUT4          : out    std_logic
 );
end component;


begin  -- rtl

	syscon.clk <= sys_clk;
        
---------------------------------------------------
--  Clock management and GPIF clock generation
---------------------------------------------------

  IBUFG_inst_clk100 : IBUFG
  generic map (
	 IBUF_LOW_PWR => TRUE, -- Low power (TRUE) vs. performance (FALSE) setting for referenced I/O standards
	 IOSTANDARD => "DEFAULT")
  port map (
	 O => clk100_i, -- Clock buffer output
	 I => pin_clk100_i	-- Clock buffer input (connect directly to top-level port)
  );

      
gpif_clock_generator : gpif_clock
  port map
   (-- Clock in ports
    CLK_IN1 => clk100_i,
    --Clock out ports
    CLK_OUT1 => clk_gpif,   -- 100.000 MHz
    CLK_OUT2 => open,       -- 150.000 MHz
    CLK_OUT3 => sys_clk,     -- 180.000 MHz
    CLK_OUT4 => open     -- 50.000 MHz
    );
   
-- produce a 180° version of the clock	
	clk_gpif_n <= not clk_gpif;

-- use the clock and its 180° version with a ODDR2 to output the clock on a regular I/O pin.	
  inst_ODDR2_clk : ODDR2
  generic map(
	DDR_ALIGNMENT => "C0",			-- Sets output alignment to "NONE", "C0", "C1" : Input data D0, D1 is alligned to C0 = i_clk
	INIT		  => '0',			-- Sets initial state of the Q output to '0' or '1'
	SRTYPE		  => "ASYNC")		-- Specifies "SYNC" or "ASYNC" set/reset
  port map (						
	Q			  => gpifclk_o,	-- 1-bit output data
	C0			  => clk_gpif,		-- 1-bit clock input
	C1			  => clk_gpif_n,	-- 1-bit clock input
	CE			  => '1',			-- 1-bit clock enable input
	D0			  => '1',			-- 1-bit data input (associated with C0)
	D1			  => '0',			-- 1-bit data input (associated with C1)
	R			  => '0',			-- 1-bit reset input
	S			  => '0'			-- 1-bit set input
  );

pin_gpifclk_o <= gpifclk_o;

  
-------------------------------------------------------
-- produce sys_clock to drive the user fpga-fabric  
-------------------------------------------------------   
    inst_clock_alignment : clock_alignment
	 port map
   (-- Clock in ports
	CLK_IN1 => pin_gpifclk_i,
	CLK_OUT1 => fx3_plclk,
	RESET	=> '0',
	LOCKED => sys_clk_locked
    );    
   
   
  process(syscon)
  begin
    if syscon.clk'event and syscon.clk = '1' then
        sysreset_i <= pin_sysreset_i; -- sync external Reset signal
        syscon.rst <= sysreset_i or (not sys_clk_locked);
    end if;
  end process;
    
  
  	-----------------------------------------------------------------------------------------------
	-- Blink the Led !
	-----------------------------------------------------------------------------------------------

    ledclk <= clk_gpif;
    
	process (ledclk)
		variable cnt: INTEGER RANGE 0 to BlinkingLedToggleCycles;
	begin
		if rising_edge (ledclk) then
            -- if (syscon.rst = '1') then
                -- BlinkingLed <= '0';
                -- cnt := 0;
            -- else    
                if (cnt = 0) then
                    cnt := BlinkingLedToggleCycles ;
                    BlinkingLed <= not BlinkingLed;
                else
                    cnt := cnt - 1;
                end if;
--            end if;
		end if;
	end process;

	pin_user_led_o <= BlinkingLed;
 

  inst_wb_intercon : wb_intercon
	generic map (
	  NR_OF_DBGPORTS => 1
	)
	port map (
	  syscon_i => syscon,

	  master_i => intercon.masters.master,
	  slave_o  => intercon.masters.slave,

	  master_o => intercon.slaves.master,
	  slave_i  => intercon.slaves.slave,

	  debug => open
	);
	
  inst_wb_ma_fx3 : wb_ma_fx3
	generic map (
	  NR_OF_DBGPORTS => 16
	)
	port map (
	  syscon_i => syscon,
	  master_o => intercon.masters.master(MA_fx3_ID),
	  slave_i  => intercon.masters.slave(MA_fx3_ID),
      
      pin_sysreset_i => pin_sysreset_i,
      
      fx3_plclk      => fx3_plclk,
	  fx3_flag_i	 => pin_fx3_flag_i,
	  fx3_slcs_n_o   => pin_fx3_slcs_n_o,
	  fx3_sloe_n_o	 => pin_fx3_sloe_n_o,
	  fx3_slrd_n_o	 => pin_fx3_slrd_n_o,
	  fx3_slwr_n_o	 => pin_fx3_slwr_n_o,
	  fx3_pktend_n_o => pin_fx3_pktend_n_o,
	  fx3_fifoadr_o	 => pin_fx3_fifoadr_o,
	  fx3_fd_io		 => pin_fx3_fd_io,

--	  debug => pin_debug_o
	  debug => open
	);
 
  inst_wb_sl_bram : wb_sl_bram
	generic map (
	  NR_OF_DBGPORTS => 1
	)
	port map (
	  syscon_i => syscon,
	  master_i => intercon.slaves.master(SL_BRAM_ID),
	  slave_o  => intercon.slaves.slave(SL_BRAM_ID),
	  
	  debug => open
	);

end rtl;
