--------------------------------------------------------------------------------
-- copyright notice:
--------------------------------------------------------------------------------
-- copyright (c) 2010 by
--
-- CESYS GmbH / GERMANY
--
-- http://www.cesys.com
--------------------------------------------------------------------------------
-- author:
--------------------------------------------------------------------------------
-- Stefan Fischer <Stefan.Fischer@cesys.com>
--------------------------------------------------------------------------------
-- license:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS NOT FREE! IT IS FOR USE TOGETHER WITH CESYS PRODUCTS ONLY!
-- YOU ARE NOT ALLOWED TO MODIFY AND DISTRIBUTE OR USE IT WITH ANY OTHER
-- HARDWARE, SOFTWARE OR ANY OTHER KIND OF ASIC OR PROGRAMMABLE LOGIC DESIGN
-- WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDER!
--------------------------------------------------------------------------------
-- disclaimer of warranty:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
-- THERE IS NO WARRANTY OR SUPPORT FOR THIS SOURCECODE. THE COPYRIGHT HOLDER 
-- PROVIDES THIS SOURCECODE "AS IS" WITHOUT WARRANTY OF ANY KIND,
-- EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
-- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-- THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THIS SOURCECODE IS WITH
-- YOU. SHOULD THIS SOURCECODE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
-- NECESSARY SERVICING, REPAIR OR CORRECTION.
--
-- IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES, INCLUDING
-- ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
-- USE OR INABILITY TO USE THIS SOURCECODE (INCLUDING BUT NOT LIMITED TO LOSS OF
-- DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD 
-- PARTIES OR A FAILURE OF THIS SOURCECODE TO OPERATE WITH ANY OTHER
-- SOFTWARE-PROGRAMS, HARDWARE-CIRCUITS OR ANY OTHER KIND OF ASIC OR
-- PROGRAMMABLE LOGIC DESIGN), EVEN IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF 
-- THE POSSIBILITY OF SUCH DAMAGES.
--------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

use work.wishbone_pkg.all;


entity wb_intercon is
  generic
  (
    NR_OF_DBGPORTS : positive := 1
  );
  port
  (
    syscon_i : in rec_syscon_port;

    master_i : in arr_master_port(RNG_MASTERS);
    slave_o : out arr_slave_port(RNG_MASTERS);

    master_o : out arr_master_port(RNG_SLAVES);
    slave_i : in arr_slave_port(RNG_SLAVES);

    debug : out std_logic_vector((NR_OF_DBGPORTS-1) downto 0)
  );
end wb_intercon;


architecture rtl of wb_intercon is

  signal intercon : rec_intercon_signal := INTERCON_DEFAULT;
  
  signal syscon : rec_syscon_port := SYSCON_DEFAULT;
  signal master : rec_master_port := MASTER_DEFAULT;
  signal slave : rec_slave_port := SLAVE_DEFAULT;
  
  signal rst : std_logic := '1';
  signal clk : std_logic := '1';
	
  constant ARBITER_WIDTH : positive := natural(RealMax(1.0, Ceil(Log2(real(NR_OF_MASTERS)))));
  signal arbiter : unsigned((ARBITER_WIDTH-1) downto 0) := (others => '0');

  signal ma_idx : natural := 0;

begin

  -- connect ports to internal signals
  ------------------------------------

  syscon <= syscon_i;

  intercon.masters.master <= master_i;
  slave_o <= intercon.masters.slave;

  master_o <= intercon.slaves.master;
  intercon.slaves.slave <= slave_i;

  debug <=
  (
    others => '0'
  );

  -- architecture implementation
  ------------------------------
  
  rst <= syscon.rst;
  clk <= syscon.clk;
  
  ma_idx <= To_Integer(arbiter);
  
  -- building shared bus topology
  process(clk, intercon, ma_idx, master, slave)
    variable k : natural := 0;
  begin
	
    -- clocked process part
    -----------------------

    -- arbitration
    if clk'event and clk = '1' then
      
      for i in RNG_MASTERS loop
        -- make sure not to take bus away from masters, which currently transfers data
        if (i = ma_idx) and intercon.masters.master(ma_idx).cyc = '0' then
          k := i;
          for dnc in 0 to natural(RealMax(0.0, real(nr_of_masters-2))) loop
            if k = (nr_of_masters-1) then
              k := 0;
            else
              k := k + 1;
            end if;
            if intercon.masters.master(k).cyc = '1' then
              arbiter <= To_Unsigned(k, ARBITER_WIDTH);
              exit;
            end if;
          end loop;
        end if;
      end loop;
        
      if rst = '1' then
        arbiter <= (others => '0');
      end if;
        
    end if;

    -- combinatorial process part
    -----------------------------

    -- master mux
    master <= MASTER_DEFAULT;
    intercon.masters.slave <= (others => slave);
    -- arbiter value selects the appropriate master device
    -- the for-loop represents the master mux
    for i in RNG_MASTERS loop
      -- ack is the one and only bit in the slave signal, which has to be muxed,
      -- you could mux the whole slave signal too, but this would waste FPGA resources
      intercon.masters.slave(i).ack <= '0';
      if i = ma_idx then
        intercon.masters.slave(i).ack <= slave.ack;
        master <= intercon.masters.master(i);
      end if;
    end loop;

    -- slave mux
    slave <= SLAVE_DEFAULT;
    intercon.slaves.master <= (others => master);
    -- partial address decoding, using some msbs of address to
    -- select the appropriate slave device
    -- the for-loop represents the slave mux
    for i in RNG_SLAVES loop
      -- cyc is the one and only bit in the master signal, which has to be muxed,
      -- you could mux the whole master signal too, but this would waste FPGA resources
      intercon.slaves.master(i).cyc <= '0'; -- default: slave is deselected
      if i = To_Integer(unsigned(master.adr(RNG_SLAVE_SELECT))) then
        intercon.slaves.master(i).cyc <= master.cyc;
        slave <= intercon.slaves.slave(i);
      end if;
    end loop;
  end process;

end rtl;
