--------------------------------------------------------------------------------
-- copyright notice:
--------------------------------------------------------------------------------
-- copyright (c) 2013 by  CESYS GmbH / GERMANY
--
-- http://www.cesys.com
--------------------------------------------------------------------------------
-- author: Manfred Kraus  <Manfred.Kraus@cesys.com>
--------------------------------------------------------------------------------
-- license:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS NOT FREE! IT IS FOR USE TOGETHER WITH CESYS PRODUCTS ONLY!
-- YOU ARE NOT ALLOWED TO MODIFY AND DISTRIBUTE OR USE IT WITH ANY OTHER
-- HARDWARE, SOFTWARE OR ANY OTHER KIND OF ASIC OR PROGRAMMABLE LOGIC DESIGN
-- WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDER!
--------------------------------------------------------------------------------
-- disclaimer of warranty:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
-- THERE IS NO WARRANTY OR SUPPORT FOR THIS SOURCECODE. THE COPYRIGHT HOLDER 
-- PROVIDES THIS SOURCECODE "AS IS" WITHOUT WARRANTY OF ANY KIND,
-- EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
-- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-- THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THIS SOURCECODE IS WITH
-- YOU. SHOULD THIS SOURCECODE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
-- NECESSARY SERVICING, REPAIR OR CORRECTION.
--
-- IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES, INCLUDING
-- ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
-- USE OR INABILITY TO USE THIS SOURCECODE (INCLUDING BUT NOT LIMITED TO LOSS OF
-- DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD 
-- PARTIES OR A FAILURE OF THIS SOURCECODE TO OPERATE WITH ANY OTHER
-- SOFTWARE-PROGRAMS, HARDWARE-CIRCUITS OR ANY OTHER KIND OF ASIC OR
-- PROGRAMMABLE LOGIC DESIGN), EVEN IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF 
-- THE POSSIBILITY OF SUCH DAMAGES.
--------------------------------------------------------------------------------


-- This is the entity of the WISHBONE master, which converts the CESYS USB protocol into
-- 32 Bit single read/write WISHBONE cycles.
-- Protocol structure: 1 Header Block followed by one ore more data blocks.
-- Blocks have a fixed size of 8192 Bytes

-- Header: 
-- CMD Byte | SIZE | ADDR | Dummy 
-- Dummy:  bytes to fill 8192 byte buffer

-- CMD Byte:
-- bit 7	data direction: 1 = Host(PC) to Peripheral(FPGA), 0 = Peripheral to Host  
-- bit 6	address increment: 1 = auto increment, 0 = fixed address
-- bit 5	burst enable: 1 = transfer every clock cycle at OCB (on-chip bus) possible, 0 = bursting disabled

-- SIZE:	Payload size (24 bits) 0x000000 means 16 X 1024 x 1024 bytes
-- ADDR:	Start address (32 bits)


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.wishbone_pkg.all;


entity wb_ma_fx3 is
  generic
  (
    NR_OF_DBGPORTS : positive := 1
  );
  port
  (
    syscon_i : in rec_syscon_port;
    master_o : out rec_master_port;
    slave_i : in rec_slave_port;

    pin_sysreset_i : in  std_logic;
	
    fx3_plclk : in std_logic;
    fx3_flag_i : in std_logic_vector(3 downto 0);
    fx3_slcs_n_o : out std_logic;
    fx3_sloe_n_o : out std_logic;
    fx3_slrd_n_o : out std_logic;
    fx3_slwr_n_o : out std_logic;
    fx3_pktend_n_o : out std_logic;
    fx3_fifoadr_o : out std_logic_vector(1 downto 0);
    fx3_fd_io : inout std_logic_vector(31 downto 0);

    debug : out std_logic_vector((NR_OF_DBGPORTS-1) downto 0)
  );
end wb_ma_fx3;


architecture rtl of wb_ma_fx3 is

  signal syscon : rec_syscon_port := SYSCON_DEFAULT;
  signal master : rec_master_port := MASTER_DEFAULT;
  signal slave : rec_slave_port := SLAVE_DEFAULT;

  signal fx3_flag_flopped : std_logic_vector(3 downto 0) :=  "0101";
  attribute IOB : string;
  attribute IOB of fx3_flag_flopped : signal is "force";
  
  signal fx3_slcs_n_flopped : std_logic := '1';
  signal fx3_sloe_n_flopped : std_logic := '1';
  signal fx3_slrd_n_flopped : std_logic := '1';
  signal fx3_slwr_n_flopped : std_logic := '1';
  signal fx3_fd_input_flopped : std_logic_vector (31 downto 0) := (others => '0');
  signal fx3_fd_output : std_logic_vector (31 downto 0) := (others => '0');
  signal fx3_fifoadr_flopped : std_logic_vector(1 downto 0) := (others => '0');
  
  
  signal fx3_flag : std_logic_vector(3 downto 0) := "0101";
  signal fx3_sloe_n : std_logic := '1';
  signal fx3_slcs_n : std_logic := '1';
  signal fx3_slrd_n : std_logic := '1';
  signal fx3_slwr_n : std_logic := '1';
  signal fx3_pktend_n : std_logic := '1';
  signal fx3_fifoadr : std_logic_vector(1 downto 0) := (others => '0');

  constant INPUT : std_logic := '0';
  constant OUTPUT : std_logic := not INPUT;
  type rec_fx3_fd is record
    oe : std_logic;
    output_flopped : std_logic_vector(31 downto 0);
    input : std_logic_vector(31 downto 0);
  end record;
  constant fx3_fd_default : rec_fx3_fd :=
  (
    oe => INPUT,
    output_flopped => (others => '0'),
    input => (others => '0')
  );
  signal fx3_fd : rec_fx3_fd := fx3_fd_default;
  
  COMPONENT Async_Fifo_8k0x32
  PORT (
    rst : IN STD_LOGIC;
    wr_clk : IN STD_LOGIC;
    rd_clk : IN STD_LOGIC;
    din : IN STD_LOGIC_VECTOR(31 DOWNTO 0);
    wr_en : IN STD_LOGIC;
    rd_en : IN STD_LOGIC;
    dout : OUT STD_LOGIC_VECTOR(31 DOWNTO 0);
    full : OUT STD_LOGIC;
    empty : OUT STD_LOGIC;
    rd_data_count : OUT STD_LOGIC_VECTOR(12 DOWNTO 0);
    wr_data_count : OUT STD_LOGIC_VECTOR(12 DOWNTO 0)
  );
END COMPONENT;

	signal gpif_rst : std_logic := '1';
	signal clk : std_logic := '0';

	signal h2p_fifo_dout : std_logic_vector(31 downto 0) := (others => '0');
	signal h2p_fifo_count : std_logic_vector(12 downto 0) := (others => '0');
    signal h2p_fifo_wr_en: std_logic := '0';	
    signal h2p_fifo_rd_en : std_logic := '0';
    signal h2p_fifo_empty : std_logic := '0';

	signal p2h_fifo_count : std_logic_vector(12 downto 0) := (others => '0');
    signal p2h_fifo_wr_en : std_logic := '0';
    signal p2h_fifo_rd_en : std_logic := '0';
    signal p2h_fifo_full : std_logic := '0';
  
    signal h2p_dval : std_logic := '0';
	signal h2p_want_rd : std_logic := '0';
	signal p2h_want_wr : std_logic := '0';
	signal addr_inc : std_logic := '0';
    
    signal transfer_count : unsigned(21 downto 0) := (others => '0');
    signal p2h_count : unsigned(10 downto 0) := (others => '0');
    signal h2p_count : unsigned(10 downto 0) := (others => '0');
  
  type t_states is
  (
    S_DECODE,
    S_ADDRESS,
    S_USB2WB,
    S_USB2WB_ACK,
    S_USB2WB_DUMMY,
    S_WB2USB,
    S_WB2USB_ACK,
    S_WB2USB_DUMMY,
    S_HEADER_DUMMY
  );
  signal state : t_states := S_DECODE;

  

  type t_states_gpif is 
  (
	S_GPIF_IDLE,
	S_H2P_PREPARE_0, 
	S_H2P_PREPARE_1, 
	S_H2P_PREPARE_2, 
	S_H2P_MOVE_DATA, 
	S_P2H_PREPARE_0, 
	S_P2H_PREPARE_1, 
	S_P2H_PREPARE_2, 
	S_P2H_MOVE_DATA,    
	S_RECOVERY_0,
	S_RECOVERY_1

  );
  
	signal state_gpif : t_states_gpif := S_GPIF_IDLE;

	signal rst : std_logic := '1';

	signal h2p_possible : std_logic := '0';
	signal p2h_possible : std_logic := '0';

	signal pipe: std_logic_vector (3 downto 0); -- mit 2 gehts bei 10 MHz !!!
	signal gpif_transaction_counter: unsigned(11 downto 0):= (others => '0');
	signal gpif_transaction_in_progress: std_logic := '0';
	signal start_gpif_transaction: std_logic := '0';
        
    constant FX3_POP_IS_EMPTY : natural := 0; -- FLAGA
    constant FX3_POP_CAN_GIVE_A_BUFFER : natural := 1; -- FLAGB
    constant FX3_PUSH_IS_FULL : natural := 2; -- FLAGC
    constant FX3_PUSH_BUFFER_POSSIBLE : natural := 3;  -- FLAGD
        
begin

	-- ports of WISHBONE bus
	syscon <= syscon_i;
	master_o <= master;
	slave <= slave_i;

   debug <=
   (
     others => '0'
   );

process(fx3_plclk, pin_sysreset_i)
    begin
        if fx3_plclk'event and fx3_plclk = '1' then 
            gpif_rst <= pin_sysreset_i;
		end if;
	end process;
    
  clk <= syscon.clk;
  rst <= syscon.rst;
	
  -- architecture implementation
  ------------------------------

  -- cyc is sometimes used for arbitration purposes,
  -- so this signal should not be statically set to '1'
    master.cyc <= master.stb;
    master.sel <= (others => '1');
    master.dat <= h2p_fifo_dout;
  
    h2p_fifo_rd_en <= h2p_want_rd and (not h2p_fifo_empty);
    p2h_fifo_wr_en <= (p2h_want_wr or (slave.ack and (not master.we))) and (not p2h_fifo_full);
  
----------------------------------------   

  SM_Fifo_to_Wishbone: process(clk,state)
  begin
    if clk'event and clk = '1' then      
      -- obligatory one clk cycle delay for fpga fifos
      h2p_dval <= h2p_fifo_rd_en;
      
      if h2p_fifo_rd_en = '1' then
        h2p_count <= h2p_count + 1; 
      end if;
      if p2h_fifo_wr_en = '1' then
        p2h_count <= p2h_count + 1; 
      end if;
      
      -- fsm decoding transfer protocol and handling wishbone interface
      case state is
        -- getting direction, size and transfer mode
        when S_DECODE =>
          if h2p_dval = '1' then
            master.we <= h2p_fifo_dout(31);
            addr_inc <= h2p_fifo_dout(30);
            -- transfer_count counts wishbone transactions not byte transfers
            -- => shift 2 bits
            transfer_count <= unsigned(h2p_fifo_dout(23 downto 2));
            state <= S_ADDRESS;
          end if;
        -- getting start address
        when S_ADDRESS => 
          if h2p_dval = '1' then
            master.adr <= h2p_fifo_dout(31 downto 2);
            state <= S_HEADER_DUMMY;
          end if;
        -- reading don't cares from pipeline
        -- data direction decision point
        when S_HEADER_DUMMY =>
        if (h2p_fifo_rd_en = '1') and (unsigned(h2p_count) = 2047) then
            if master.we = '1' then
              state <= S_USB2WB;
            else
              h2p_want_rd <= '0';
              state <= S_WB2USB;
            end if;
          end if;
        -- transfering data usb => wishbone
        when S_USB2WB =>
          if h2p_fifo_rd_en = '1' then
            h2p_want_rd <= '0';
            master.stb <= '1';
            state <= S_USB2WB_ACK;
          end if;
        when S_USB2WB_ACK =>
          if slave.ack = '1' then
            h2p_want_rd <= '1';
            
            master.stb <= '0';
            transfer_count <= transfer_count - 1;
            if addr_inc = '1' then
              master.adr <= std_logic_vector(unsigned(master.adr) + 1);
            end if;
            if transfer_count = 1 then
              if unsigned(h2p_count) = 0 then
                state <= S_DECODE;
              else
                state <= S_USB2WB_DUMMY;
              end if;
            else
              state <= S_USB2WB;
            end if;
          end if;
        -- reading dummy data from pipeline
        when S_USB2WB_DUMMY =>
          if (h2p_dval = '1') and (unsigned(h2p_count) = 0) then
            state <= S_DECODE;
          end if;
        -- transfering data wishbone => usb 
        when S_WB2USB =>
          if p2h_fifo_full = '0' then
            master.stb <= '1';
            state <= S_WB2USB_ACK;
          end if;
        when S_WB2USB_ACK =>
          if slave.ack = '1' then
            master.stb <= '0';
            transfer_count <= transfer_count - 1;
            if addr_inc = '1' then
              master.adr <= std_logic_vector(unsigned(master.adr) + 1);
            end if;
            if transfer_count = 1 then
              if unsigned(p2h_count) = 2047 then
                h2p_want_rd <= '1';
                state <= S_DECODE;
              else
                p2h_want_wr <= '1';
                state <= S_WB2USB_DUMMY;
              end if;
            else
              state <= S_WB2USB;
            end if;
          end if;
        -- adding dummy data to pipeline
        when S_WB2USB_DUMMY =>
          if (p2h_fifo_wr_en = '1') and (unsigned(p2h_count) = 2047) then
            h2p_want_rd <= '1';
            p2h_want_wr <= '0';
            state <= S_DECODE;
          end if;
        when others => null;
      end case;
      
      if rst = '1' then
        h2p_dval <= '0';
        h2p_want_rd <= '1';
        h2p_count <= (others => '0');
        p2h_want_wr <= '0';
        p2h_count <= (others => '0');
        state <= S_DECODE;
        master.stb <= '0';
      end if;
      
    end if;
  end process;

-- Host to Peripheral  
  H2P_FIFO : Async_Fifo_8k0x32
  PORT MAP (
    rst => gpif_rst,
    wr_clk => fx3_plclk,
    rd_clk => clk,
    din => fx3_fd_input_flopped,
    wr_en => h2p_fifo_wr_en,
    rd_en => h2p_fifo_rd_en,
    dout => h2p_fifo_dout,
    full => open,
    empty => h2p_fifo_empty,
    rd_data_count => open,
    wr_data_count => h2p_fifo_count
  );
  
  
-- Peripheral to Host
  P2H_FIFO : Async_Fifo_8k0x32
  PORT MAP (
    rst => gpif_rst,
    wr_clk => clk,
    rd_clk => fx3_plclk,
    din => slave.dat,
    wr_en => p2h_fifo_wr_en,
    rd_en => p2h_fifo_rd_en,
    dout => fx3_fd_output,
    full => p2h_fifo_full,
    empty => open,
    rd_data_count => p2h_fifo_count,
    wr_data_count => open
  );

  
  
-- -----------------------------------------------------------------------
--  connect H2P and P2H FIFOs to FX-3 slave FiFo using GPIF-II interface  
-- ------------------------------------------------------------------------
 
 
 -- dataflow is controlled by FX-3 flag signals
 --
 -- *** SOCKET 0 / THREAD 0
 --   Data direction: FX3 to FPGA (H2P)  "POP FX-3"
 --   FlagA (active high, initial high): thread0 DMA ready, Empty  		'1' means "FX3 can POP nothing"
 --   FlagB (active low, initial low): thread0 Watermark(1023)			'1' means "FX3 can POP a buffer"
 --
 -- *** SOCKET 2 / THREAD 2
 --   Data direction: FPGA to FX3 (P2H)  "PUSH FX-3"
 --   FlagC (active high, initial high): thread2 DMA ready, Full  		'1' means "FX3 no PUSH possible because it is full"
 --   FlagD (active high, initial low): thread0 Watermark(15*1024)-1	'1' means "FX3 PUSH a buffer is possible"
 --
 -- Watermark Flags are only valid in conjunction with the full/empty flag.
 --
    
  fx3_fifoadr(0) <= '0'; -- fixed, because only two fx3 sockets used (socket0 out(h2p) a[1:0]=b"00", socket2 in(p2h) a[1:0]=b"10")
  fx3_pktend_n <= '1'; -- no short packet transfer, the usb packet size is fixed at 8 * 1024 bytes, !must! be tied to '1'

  process(fx3_fd, fx3_fd_io)
  begin
    if fx3_fd.oe = OUTPUT then
      fx3_fd_io <= fx3_fd.output_flopped;
    else
      fx3_fd_io <= (others => 'Z');
    end if;
  end process;
  fx3_fd.input <= fx3_fd_io;


-- flop all fx3 input signals to ensure perfect timing
  process(fx3_plclk, fx3_fd.input, fx3_flag)
  	begin
		if fx3_plclk'event and fx3_plclk = '1' then 
				fx3_fd_input_flopped <= fx3_fd.input;
				fx3_flag_flopped <= fx3_flag;
		end if;
	end process;


-- flop all fx3 output signals to ensure perfect timing
  process(fx3_plclk, gpif_rst, fx3_fd_output, fx3_sloe_n, fx3_slcs_n,
			fx3_slrd_n, fx3_slwr_n, fx3_fifoadr)
	begin
		if fx3_plclk'event and fx3_plclk = '1' then 
			fx3_fd.output_flopped <= fx3_fd_output;
			fx3_slcs_n_flopped <= fx3_slcs_n;  
			fx3_sloe_n_flopped <= fx3_sloe_n;  
			fx3_slrd_n_flopped <= fx3_slrd_n;
			fx3_slwr_n_flopped <= fx3_slwr_n;
			fx3_fifoadr_flopped <= fx3_fifoadr;		
		end if;

		if gpif_rst = '1' then
			fx3_slcs_n_flopped <= 'Z';  
			fx3_sloe_n_flopped <= 'Z';  
			fx3_slrd_n_flopped <= 'Z';
			fx3_slwr_n_flopped <= 'Z';
			fx3_fifoadr_flopped <= "ZZ";
		end if;	
	end process;
	
	
   -- ports of slave fifo interface
  fx3_flag <= fx3_flag_i;
  fx3_slcs_n_o <= fx3_slcs_n_flopped;
  fx3_sloe_n_o <= fx3_sloe_n_flopped;
  fx3_slrd_n_o <= fx3_slrd_n_flopped;
  fx3_slwr_n_o <= fx3_slwr_n_flopped;
  fx3_pktend_n_o <= fx3_pktend_n;
  fx3_fifoadr_o <= fx3_fifoadr_flopped;
       
    p2h_possible <= (fx3_flag_flopped(FX3_PUSH_BUFFER_POSSIBLE))
	and (not fx3_flag_flopped(FX3_PUSH_IS_FULL))
    and '1' when (unsigned(p2h_fifo_count) > 2047) else '0'; -- more than 2047 words (8 kByte) available

    h2p_possible <= (fx3_flag_flopped(FX3_POP_CAN_GIVE_A_BUFFER))
	and (not fx3_flag_flopped(FX3_POP_IS_EMPTY))
--    and '1' when(to_integer(unsigned(h2p_fifo_count)) < (8192 - 2048)) else '0'; -- space for 2048 words (8 kByte) left
    and '1' when(to_integer(unsigned(h2p_fifo_count)) < (8191 - 2048)) else '0'; -- space for 2048 words (8 kByte) left
                
	-- Generate FX3 write strobe
	process (fx3_plclk, p2h_fifo_rd_en)
	begin
		if fx3_plclk'event and fx3_plclk = '1' then
			if gpif_rst = '1' then
				fx3_slwr_n <= '1';
			else		
				fx3_slwr_n <= not p2h_fifo_rd_en;
			end if;
		end if;
	end process;
					
	-- Generate H2P Fifo Write strobe					
		-- 4 clocks delay: +1clk: flopping-out slrd +2clk fx3 data latency +1clk flopping-in fx3 data
	process (fx3_plclk, gpif_rst, pipe, fx3_slrd_n)
	begin
		if fx3_plclk'event and fx3_plclk = '1' then
			if gpif_rst = '1' then
				h2p_fifo_wr_en <= '0';
				pipe <= (others => '1');
			else		
				pipe <= pipe((pipe'high)-1 downto 0) & fx3_slrd_n; -- 3 delay pipe + this one = 4
			end if;
		end if;
		h2p_fifo_wr_en <= not pipe(pipe'high);
	end process;
			

	-- each gpif transaction has a fixed block length of 8192 bytes = 2048 words.
    process (fx3_plclk, gpif_rst)
    begin
        if (fx3_plclk'event and fx3_plclk='1') then
			if (gpif_rst='1') then
				gpif_transaction_counter <= (others => '0');
			else 
				if (start_gpif_transaction = '1') then
					gpif_transaction_counter <= "100000000000"; -- 2048 (8k Byte)
				else
					if gpif_transaction_counter /= 0 then
						gpif_transaction_counter <= gpif_transaction_counter - 1;
					end if;
				end if;
			end if;
		end if;
    end process;

	
gpif_transaction_in_progress <= '0' when gpif_transaction_counter = "000000000000" else '1';
fx3_slrd_n <= not (gpif_transaction_in_progress and (not fx3_fifoadr(1)));	
p2h_fifo_rd_en <= gpif_transaction_in_progress and fx3_fifoadr(1);

			
	SM_GPIF_to_Fifo: process ( fx3_plclk, gpif_rst, state_gpif, h2p_possible, p2h_possible, gpif_transaction_in_progress)
	begin    
		if fx3_plclk'event and fx3_plclk = '1' then

		start_gpif_transaction <= '0';

		if gpif_rst = '1' then
			state_gpif <= S_GPIF_IDLE;
		else
			case state_gpif is

				when S_GPIF_IDLE =>
					fx3_fd.oe <= INPUT;
					fx3_fifoadr(1) <= '0';
					fx3_sloe_n <= '1';
					fx3_slcs_n <= '1';
					if (h2p_possible = '1') then
						state_gpif <= S_H2P_PREPARE_0;
					else 
						if (p2h_possible = '1') then
							state_gpif <= S_P2H_PREPARE_0;
						end if;
					end if;
					
				when S_H2P_PREPARE_0 =>
					fx3_slcs_n <= '0';
					fx3_fd.oe <= INPUT;		-- FPGA listens to data-bus 
					fx3_fifoadr(1) <= '0'; 	-- h2p uses socket 0
					state_gpif <= S_H2P_PREPARE_1;

				when S_H2P_PREPARE_1 =>
					fx3_sloe_n <= '0';		-- fx3 drives data-bus 
					start_gpif_transaction <= '1';
					state_gpif <= S_H2P_PREPARE_2;

				when S_H2P_PREPARE_2 =>
					state_gpif <= S_H2P_MOVE_DATA;

				when S_H2P_MOVE_DATA =>
					if (gpif_transaction_in_progress = '0') then
--						fx3_slrd_n <= '1';
						fx3_slcs_n <= '1';		
						state_gpif <= S_RECOVERY_0;
					end if;
				
				when S_P2H_PREPARE_0 =>
					fx3_slcs_n <= '0';
					fx3_sloe_n <= '1';		-- fx3 listens to data-bus 
					fx3_fifoadr(1) <= '1'; 	-- h2p uses socket 2
					state_gpif <= S_P2H_PREPARE_1;

				when S_P2H_PREPARE_1 =>
					start_gpif_transaction <= '1';
					fx3_fd.oe <= OUTPUT;	-- FPGA drives data-bus 
					state_gpif <= S_P2H_PREPARE_2;

				when S_P2H_PREPARE_2 =>
					state_gpif <= S_P2H_MOVE_DATA;

				when S_P2H_MOVE_DATA =>
					if (gpif_transaction_in_progress = '0') then
						fx3_slcs_n <= '1';		
						state_gpif <= S_RECOVERY_0;
					end if;

                -- time to finish transaction (pipelined signals!)
				when S_RECOVERY_0 =>
					state_gpif <= S_RECOVERY_1;

                when S_RECOVERY_1 =>
					state_gpif <= S_GPIF_IDLE;
					
				when others => null;

				end case;
		end if;
    end if;

  end process;  
  
end rtl;
