--------------------------------------------------------------------------------
-- copyright notice:
--------------------------------------------------------------------------------
-- copyright (c) 2010 by
--
-- CESYS GmbH / GERMANY
--
-- http://www.cesys.com
--------------------------------------------------------------------------------
-- author:
--------------------------------------------------------------------------------
-- Stefan Fischer <Stefan.Fischer@cesys.com>
--------------------------------------------------------------------------------
-- license:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS NOT FREE! IT IS FOR USE TOGETHER WITH CESYS PRODCUTS ONLY!
-- YOU ARE NOT ALLOWED TO MODIFY AND DISTRIBUTE OR USE IT WITH ANY OTHER
-- HARDWARE, SOFTWARE OR ANY OTHER KIND OF ASIC OR PROGRAMMABLE LOGIC DESIGN
-- WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDER!
--------------------------------------------------------------------------------
-- disclaimer of warranty:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
-- THERE IS NO WARRANTY OR SUPPORT FOR THIS SOURCECODE. THE COPYRIGHT HOLDER 
-- PROVIDES THIS SOURCECODE "AS IS" WITHOUT WARRANTY OF ANY KIND,
-- EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
-- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-- THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THIS SOURCECODE IS WITH
-- YOU. SHOULD THIS SOURCECODE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
-- NECESSARY SERVICING, REPAIR OR CORRECTION.
--
-- IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES, INCLUDING
-- ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
-- USE OR INABILITY TO USE THIS SOURCECODE (INCLUDING BUT NOT LIMITED TO LOSS OF
-- DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD 
-- PARTIES OR A FAILURE OF THIS SOURCECODE TO OPERATE WITH ANY OTHER
-- SOFTWARE-PROGRAMS, HARDWARE-CIRCUITS OR ANY OTHER KIND OF ASIC OR
-- PROGRAMMABLE LOGIC DESIGN), EVEN IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF 
-- THE POSSIBILITY OF SUCH DAMAGES.
--------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;

use work.wishbone_pkg.all;
use work.wb_sl_gpio_pkg.all;


entity wb_sl_gpio is
  generic
  (
    NR_OF_DBGPORTS : positive := 1
  );
  port
  (
    syscon_i : in rec_syscon_port;
    master_i : in rec_master_port;
    slave_o : out rec_slave_port;
    
    gpio_output_o : out std_logic_vector(255 downto 0);
    gpio_oe_o : out std_logic_vector(255 downto 0);
    gpio_input_i : in std_logic_vector(255 downto 0);

    debug : out std_logic_vector((NR_OF_DBGPORTS-1) downto 0)
  );
end wb_sl_gpio;


architecture rtl of wb_sl_gpio is

  signal syscon : rec_syscon_port := SYSCON_DEFAULT;
  signal master : rec_master_port := MASTER_DEFAULT;
  signal slave : rec_slave_port := SLAVE_DEFAULT;

  signal rst : std_logic := '1';
  signal clk : std_logic := '1';
  
  signal wb_reg_we : std_logic := '0';

  signal gpio_output : std_logic_vector(255 downto 0) := (others => '0');
  signal gpio_oe     : std_logic_vector(255 downto 0) := (others => INPUT);
  signal gpio_input  : std_logic_vector(255 downto 0) := (others => '0');
  signal gpi_ff : std_logic_vector(255 downto 0) := (others => '0');
  
begin

  -- connect ports to internal signals
  ------------------------------------
  
  -- ports of WISHBONE bus
  syscon <= syscon_i;
  master <= master_i;
  slave_o <= slave;

  -- use IOB FFs for signals of gpio
  process(clk, gpio_output, gpio_oe, gpio_input_i)
  begin
		if clk'event and clk = '1' then
			gpio_output_o <= gpio_output;
			gpio_oe_o <= gpio_oe;
			gpio_input <= gpio_input_i;
		end if;
  end process;
  
  debug <=
  (
    others => '0'
  );

  -- architecture implementation
  ------------------------------

  rst <= syscon.rst;
  clk <= syscon.clk;

  wb_reg_we <= master.cyc and master.stb and master.we;

  -- WISHBONE handshake and external 256-bit GPIO
  -- input-, output- and output-enable-registers
  process(clk)
  begin
    if clk'event and clk = '1' then
      
      -- WISHBONE handshake acknowledge response for registers
      slave.ack <= master.cyc and master.stb and (not slave.ack);
      -- default data bus value
      slave.dat <= (others => '0');
      
      -- WISHBONE address-decoding
      case master.adr(RNG_GPIO_ADR) is
        -- input/output registers
        when GPIO_IO_0_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_output(31 downto 0) <= master.dat;
          end if;
          slave.dat <= gpi_ff(31 downto 0);
        when GPIO_IO_1_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_output(63 downto 32) <= master.dat;
          end if;
          slave.dat <= gpi_ff(63 downto 32);
        when GPIO_IO_2_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_output(95 downto 64) <= master.dat;
          end if;
          slave.dat <= gpi_ff(95 downto 64);
        when GPIO_IO_3_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_output(127 downto 96) <= master.dat;
          end if;
          slave.dat <= gpi_ff(127 downto 96);
        when GPIO_IO_4_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_output(159 downto 128) <= master.dat;
          end if;
          slave.dat <= gpi_ff(159 downto 128);
        when GPIO_IO_5_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_output(191 downto 160) <= master.dat;
          end if;
          slave.dat <= gpi_ff(191 downto 160);
        when GPIO_IO_6_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_output(223 downto 192) <= master.dat;
          end if;
          slave.dat <= gpi_ff(223 downto 192);
        when GPIO_IO_7_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_output(255 downto 224) <= master.dat;
          end if;
          slave.dat <= gpi_ff(255 downto 224);
        -- output-enable registers
        when GPIO_OE_0_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_oe(31 downto 0) <= master.dat;
          end if;
          slave.dat <= gpio_oe(31 downto 0);
        when GPIO_OE_1_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_oe(63 downto 32) <= master.dat;
          end if;
          slave.dat <= gpio_oe(63 downto 32);
        when GPIO_OE_2_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_oe(95 downto 64) <= master.dat;
          end if;
          slave.dat <= gpio_oe(95 downto 64);
        when GPIO_OE_3_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_oe(127 downto 96) <= master.dat;
          end if;
          slave.dat <= gpio_oe(127 downto 96);
        when GPIO_OE_4_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_oe(159 downto 128) <= master.dat;
          end if;
          slave.dat <= gpio_oe(159 downto 128);
        when GPIO_OE_5_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_oe(191 downto 160) <= master.dat;
          end if;
          slave.dat <= gpio_oe(191 downto 160);
        when GPIO_OE_6_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_oe(223 downto 192) <= master.dat;
          end if;
          slave.dat <= gpio_oe(223 downto 192);
        when GPIO_OE_7_OFFSET(RNG_GPIO_ADR) =>
          if wb_reg_we = '1' then
            gpio_oe(255 downto 224) <= master.dat;
          end if;
          slave.dat <= gpio_oe(255 downto 224);
        when others => null;
      end case;
      
      -- route inputs through flip-flops
      gpi_ff <= gpio_input;
      
	  
      if rst = '1' then
        slave.ack <= '0';
--MK        gpio_oe <= (others => INPUT);
--MK Testen ob das schlechte timing vom asynchronen reset kommt.
      end if;
    
    end if;
  end process;
  
end rtl;
