--------------------------------------------------------------------------------
-- copyright notice:
--------------------------------------------------------------------------------
-- copyright (c) 2010 by
--
-- CESYS GmbH / GERMANY
--
-- http://www.cesys.com
--------------------------------------------------------------------------------
-- author:
--------------------------------------------------------------------------------
-- Stefan Fischer <Stefan.Fischer@cesys.com>
--------------------------------------------------------------------------------
-- license:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS NOT FREE! IT IS FOR USE TOGETHER WITH CESYS PRODUCTS ONLY!
-- YOU ARE NOT ALLOWED TO MODIFY AND DISTRIBUTE OR USE IT WITH ANY OTHER
-- HARDWARE, SOFTWARE OR ANY OTHER KIND OF ASIC OR PROGRAMMABLE LOGIC DESIGN
-- WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDER!
--------------------------------------------------------------------------------
-- disclaimer of warranty:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
-- THERE IS NO WARRANTY OR SUPPORT FOR THIS SOURCECODE. THE COPYRIGHT HOLDER 
-- PROVIDES THIS SOURCECODE "AS IS" WITHOUT WARRANTY OF ANY KIND,
-- EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
-- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-- THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THIS SOURCECODE IS WITH
-- YOU. SHOULD THIS SOURCECODE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
-- NECESSARY SERVICING, REPAIR OR CORRECTION.
--
-- IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES, INCLUDING
-- ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
-- USE OR INABILITY TO USE THIS SOURCECODE (INCLUDING BUT NOT LIMITED TO LOSS OF
-- DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD 
-- PARTIES OR A FAILURE OF THIS SOURCECODE TO OPERATE WITH ANY OTHER
-- SOFTWARE-PROGRAMS, HARDWARE-CIRCUITS OR ANY OTHER KIND OF ASIC OR
-- PROGRAMMABLE LOGIC DESIGN), EVEN IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF 
-- THE POSSIBILITY OF SUCH DAMAGES.
--------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;

use work.wishbone_pkg.all;


entity wb_sl_mcb is
  generic
  (
    NR_OF_DBGPORTS : positive := 1
  );
  port
  (
    syscon_i : in rec_syscon_port;
    master_i : in rec_master_port;
    slave_o : out rec_slave_port;

    cmd_clk : out std_logic;
    cmd_en : out std_logic;
    cmd_instr : out std_logic_vector(2 downto 0);
    cmd_bl : out std_logic_vector(5 downto 0);
    cmd_byte_addr : out std_logic_vector(29 downto 0);
    cmd_empty : in std_logic;
    cmd_full : in std_logic;
    
    wr_clk : out std_logic;
    wr_en : out std_logic;
    wr_mask : out std_logic_vector(3 downto 0);
    wr_data : out std_logic_vector(31 downto 0);
    wr_full : in std_logic;
    wr_empty : in std_logic;
    wr_count : in std_logic_vector(6 downto 0);
    wr_underrun : in std_logic;
    wr_error : in std_logic;
    
    rd_clk : out std_logic;
    rd_en : out std_logic;
    rd_data : in std_logic_vector(31 downto 0);
    rd_full : in std_logic;
    rd_empty : in std_logic;
    rd_count : in std_logic_vector(6 downto 0);
    rd_overflow : in std_logic;
    rd_error : in std_logic;

    debug : out std_logic_vector((NR_OF_DBGPORTS-1) downto 0)
  );
end wb_sl_mcb;


architecture rtl of wb_sl_mcb is

  signal syscon : rec_syscon_port := SYSCON_DEFAULT;
  signal master : rec_master_port := MASTER_DEFAULT;
  signal slave : rec_slave_port := SLAVE_DEFAULT;

  signal rst : std_logic := '1';
  signal clk : std_logic := '1';

  type t_states is
  (
    S_IDLE,
    S_WR,
    S_RD,
    S_ACK
  );
  signal state : t_states := S_IDLE;
  
begin

  -- connect ports to internal signals
  ------------------------------------
  
  -- ports of WISHBONE bus
  syscon <= syscon_i;
  master <= master_i;
  slave_o <= slave;

  debug <=
  (
    others => '0'
  );

  rst <= syscon.rst;
  clk <= syscon.clk;
  
  -- architecture implementation
  ------------------------------

  cmd_clk <= clk;
  wr_clk <= clk;
  rd_clk <= clk;
  
  cmd_instr <= (0 => not master.we, others => '0');
  cmd_bl <= (others => '0');
  cmd_byte_addr(29 downto 27) <= (others => '0');
  cmd_byte_addr(26 downto 2) <= master.adr(26 downto 2);
  cmd_byte_addr(1 downto 0) <= (others => '0');

  wr_mask <= not master.sel;
  wr_data <= master.dat;

  slave.dat <= rd_data;

  process(clk)
  begin
    if clk'event and clk = '1' then
      
      cmd_en <= '0';
      wr_en <= '0';
      rd_en <= '0';

      slave.ack <= '0';

      case state is
        when S_IDLE =>
          if (master.cyc and master.stb) = '1' then
            if master.we = '1' then
              if wr_full = '0' then
                wr_en <= '1';
                state <= S_WR;                
              end if;
            else
              if cmd_full = '0' then
                cmd_en <= '1';
                state <= S_RD;
              end if;
            end if;
          end if;
        when S_WR =>
          if ((not cmd_full) and (not wr_empty)) = '1' then
            cmd_en <= '1';
            slave.ack <= '1';
            state <= S_ACK;
          end if;
        when S_RD =>
          if rd_empty = '0' then                      
            rd_en <= '1';
            slave.ack <= '1';
            state <= S_ACK;
          end if;
        when S_ACK =>
          state <= S_IDLE;
        when others => null;
      end case;

      if rst = '1' then
        state <= S_IDLE;  
      end if;
      
    end if;
  end process;
				
end rtl;
