--------------------------------------------------------------------------------
-- copyright notice:
--------------------------------------------------------------------------------
-- copyright (c) 2010 by
--
-- CESYS GmbH / GERMANY
--
-- http://www.cesys.com
--------------------------------------------------------------------------------
-- author:
--------------------------------------------------------------------------------
-- Stefan Fischer <Stefan.Fischer@cesys.com>
--------------------------------------------------------------------------------
-- license:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS NOT FREE! IT IS FOR USE TOGETHER WITH CESYS PRODUCTS ONLY!
-- YOU ARE NOT ALLOWED TO MODIFY AND DISTRIBUTE OR USE IT WITH ANY OTHER
-- HARDWARE, SOFTWARE OR ANY OTHER KIND OF ASIC OR PROGRAMMABLE LOGIC DESIGN
-- WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDER!
--------------------------------------------------------------------------------
-- disclaimer of warranty:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
-- THERE IS NO WARRANTY OR SUPPORT FOR THIS SOURCECODE. THE COPYRIGHT HOLDER 
-- PROVIDES THIS SOURCECODE "AS IS" WITHOUT WARRANTY OF ANY KIND,
-- EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
-- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-- THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THIS SOURCECODE IS WITH
-- YOU. SHOULD THIS SOURCECODE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
-- NECESSARY SERVICING, REPAIR OR CORRECTION.
--
-- IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES, INCLUDING
-- ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
-- USE OR INABILITY TO USE THIS SOURCECODE (INCLUDING BUT NOT LIMITED TO LOSS OF
-- DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD 
-- PARTIES OR A FAILURE OF THIS SOURCECODE TO OPERATE WITH ANY OTHER
-- SOFTWARE-PROGRAMS, HARDWARE-CIRCUITS OR ANY OTHER KIND OF ASIC OR
-- PROGRAMMABLE LOGIC DESIGN), EVEN IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF 
-- THE POSSIBILITY OF SUCH DAMAGES.
--------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.wishbone_pkg.all;
use work.wb_sl_uart_pkg.all;


entity wb_sl_uart is
  generic
  (
    NR_OF_DBGPORTS : positive := 1
  );
  port
  (
    syscon_i : in rec_syscon_port;
    master_i : in rec_master_port;
    slave_o : out rec_slave_port;
    
    uart_rx_i : in std_logic;
    uart_tx_o : out std_logic;
    
    debug : out std_logic_vector((NR_OF_DBGPORTS-1) downto 0)
  );
end wb_sl_uart;


architecture rtl of wb_sl_uart is

  signal syscon : rec_syscon_port := SYSCON_DEFAULT;
  signal master : rec_master_port := MASTER_DEFAULT;
  signal slave : rec_slave_port := SLAVE_DEFAULT;

  signal uart_rx : std_logic := '1';
  signal uart_tx : std_logic := '1';
  
  component xil_uart is
    generic
    (
      NR_OF_DBGPORTS : positive := 1
    );
    port
    (
      rst : in std_logic;
      clk : in std_logic;
      
      clk16x_i : in std_logic;
  
      rx_read_buffer_i : in std_logic;
      rx_data_out_o : out std_logic_vector(7 downto 0);
      rx_buffer_data_present_o : out std_logic;
      rx_buffer_full_o : out std_logic;
      rx_buffer_half_full_o : out std_logic;
      rx_serial_in_i : in std_logic;
  
      tx_write_buffer_i : in std_logic;
      tx_data_in_i : in std_logic_vector(7 downto 0);
      tx_buffer_full_o : out std_logic;
      tx_buffer_half_full_o : out std_logic;
      tx_serial_out_o : out std_logic;
  
      debug : out std_logic_vector((NR_OF_DBGPORTS-1) downto 0)
    );
  end component;
    
  signal baud_timer : unsigned(15 downto 0) := (others => '0');
  signal baud_timer_limit : std_logic_vector(15 downto 0) := BaudCalc(9600.0, 48.0E6);
  
  signal rst : std_logic := '1';
  signal clk : std_logic := '1';
  
  signal clk16x : std_logic := '0';
  
  signal rx_read_buffer : std_logic := '0';
  signal rx_data_out : std_logic_vector(7 downto 0) := (others => '0');
  signal rx_buffer_data_present : std_logic := '0';
  signal rx_buffer_full : std_logic := '0'; 
  signal rx_buffer_half_full : std_logic := '0';
  signal rx_serial_in : std_logic := '0';
  
  signal tx_write_buffer : std_logic := '0';
  signal tx_data_in : std_logic_vector(7 downto 0) := (others => '0');
  signal tx_buffer_full : std_logic := '0'; 
  signal tx_buffer_half_full : std_logic := '0';
  signal tx_serial_out : std_logic := '0';
  
begin

  -- connect ports to internal signals
  ------------------------------------
  
  -- ports of WISHBONE bus
  syscon <= syscon_i;
  master <= master_i;
  slave_o <= slave;
  
  uart_rx <= uart_rx_i;
  uart_tx_o <= uart_tx;
  
  debug <=
  (
    others => '0'
  );
  
  rst <= syscon.rst;
  clk <= syscon.clk;
  
  -- architecture implementation
  ------------------------------
  
  rx_serial_in <= uart_rx;
  uart_tx <= tx_serial_out;
  
  tx_data_in <= master.dat(UART_RXTX_DATA_FIELD);
  
  process(clk)
  begin
    if clk'event and clk = '1' then
      
      tx_write_buffer <= '0';
      rx_read_buffer <= '0';
      
      slave.ack <= '0';
      slave.dat <= (others => '0');
      
      if baud_timer = unsigned(baud_timer_limit) then
        clk16x <= '1';
        baud_timer <= (others => '0');
      else
        clk16x <= '0';
        baud_timer <= baud_timer + 1;
      end if;
      
      case master.adr(RNG_UART_ADR) is
        when UART_RXTX_OFFSET(RNG_UART_ADR) => 
          if master.we = '1' then
            slave.ack <= master.cyc and master.stb and (not slave.ack) and (not tx_buffer_full);
            tx_write_buffer <= master.cyc and master.stb and (not slave.ack) and (not tx_buffer_full);
          else
            slave.ack <= master.cyc and master.stb and (not slave.ack);
            rx_read_buffer <= master.cyc and master.stb and (not slave.ack) and rx_buffer_data_present;
          end if;
          slave.dat(UART_RXTX_RDVAL_FLAG) <= rx_buffer_data_present;
          slave.dat(UART_RXTX_DATA_FIELD) <= rx_data_out;
        when UART_STACFG_OFFSET(RNG_UART_ADR) => 
          slave.ack <= master.cyc and master.stb and (not slave.ack);
          if (master.cyc and master.stb and master.we) = '1' then
            baud_timer_limit <= master.dat(UART_STACFG_BDR_FIELD);
            baud_timer <= (others => '0');
          end if;
          slave.dat(UART_STACFG_BDR_FIELD) <= baud_timer_limit;
          slave.dat(UART_STACFG_RF_FLAG) <= rx_buffer_full;
          slave.dat(UART_STACFG_RHF_FLAG) <= rx_buffer_half_full;
          slave.dat(UART_STACFG_TF_FLAG) <= tx_buffer_full;
          slave.dat(UART_STACFG_THF_FLAG) <= tx_buffer_half_full;
        when others => null;
      end case;
      
    end if;
  end process;
  
  inst_xil_uart : xil_uart
    generic map
    (
      NR_OF_DBGPORTS => 1
      )
    port map
    (
      rst => rst,
      clk => clk,
      
      clk16x_i => clk16x,
      
      rx_read_buffer_i => rx_read_buffer,
      rx_data_out_o => rx_data_out,
      rx_buffer_data_present_o => rx_buffer_data_present,
      rx_buffer_full_o => rx_buffer_full,
      rx_buffer_half_full_o => rx_buffer_half_full,
      rx_serial_in_i => rx_serial_in,
      
      tx_write_buffer_i => tx_write_buffer,
      tx_data_in_i => tx_data_in,
      tx_buffer_full_o => tx_buffer_full,
      tx_buffer_half_full_o => tx_buffer_half_full,
      tx_serial_out_o => tx_serial_out,
      
      debug => open
      );
  
end rtl;
