--------------------------------------------------------------------------------
-- copyright notice:
--------------------------------------------------------------------------------
-- copyright (c) 2010 by
--
-- CESYS GmbH / GERMANY
--
-- http://www.cesys.com
--------------------------------------------------------------------------------
-- author:
--------------------------------------------------------------------------------
-- Stefan Fischer <Stefan.Fischer@cesys.com>
--------------------------------------------------------------------------------
-- license:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS NOT FREE! IT IS FOR USE TOGETHER WITH CESYS PRODUCTS ONLY!
-- YOU ARE NOT ALLOWED TO MODIFY AND DISTRIBUTE OR USE IT WITH ANY OTHER
-- HARDWARE, SOFTWARE OR ANY OTHER KIND OF ASIC OR PROGRAMMABLE LOGIC DESIGN
-- WITHOUT THE EXPLICIT PERMISSION OF THE COPYRIGHT HOLDER!
--------------------------------------------------------------------------------
-- disclaimer of warranty:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
-- THERE IS NO WARRANTY OR SUPPORT FOR THIS SOURCECODE. THE COPYRIGHT HOLDER 
-- PROVIDES THIS SOURCECODE "AS IS" WITHOUT WARRANTY OF ANY KIND,
-- EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
-- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-- THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THIS SOURCECODE IS WITH
-- YOU. SHOULD THIS SOURCECODE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
-- NECESSARY SERVICING, REPAIR OR CORRECTION.
--
-- IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES, INCLUDING
-- ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
-- USE OR INABILITY TO USE THIS SOURCECODE (INCLUDING BUT NOT LIMITED TO LOSS OF
-- DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD 
-- PARTIES OR A FAILURE OF THIS SOURCECODE TO OPERATE WITH ANY OTHER
-- SOFTWARE-PROGRAMS, HARDWARE-CIRCUITS OR ANY OTHER KIND OF ASIC OR
-- PROGRAMMABLE LOGIC DESIGN), EVEN IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF 
-- THE POSSIBILITY OF SUCH DAMAGES.
--------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;


package wishbone_pkg is

  -- basic system configuration
  -----------------------------

  -- nr of devices
  constant NR_OF_MASTERS : positive := 1;
  subtype RNG_MASTERS is natural range (NR_OF_MASTERS-1) downto 0;
  constant NR_OF_SLAVES : positive := 6;
  subtype RNG_SLAVES is natural range (NR_OF_SLAVES-1) downto 0;

  -- nr of address- and data-lines
  constant ADR_WIDTH : positive := 32;
  subtype RNG_ADR is natural range (ADR_WIDTH-1) downto 2;
  constant DAT_WIDTH : positive := 32;
  subtype RNG_DAT is natural range (DAT_WIDTH-1) downto 0;
  constant SEL_WIDTH : positive := 4;
  subtype RNG_SEL is natural range (SEL_WIDTH-1) downto 0;
  
  -- WISHBONE specific data types and default values
  --------------------------------------------------

  -- ports driven by WISHBONE syscon modules
  type rec_syscon_port is record
    rst : std_logic;
    clk : std_logic;
  end record;
  constant SYSCON_DEFAULT : rec_syscon_port :=
  (
    rst => '1',
    clk => '1'
  );
  type arr_syscon_port is array(natural range <>) of rec_syscon_port;

  -- ports driven by WISHBONE master modules
  type rec_master_port is record
    cyc : std_logic;
    stb : std_logic;
    we : std_logic;
    sel : std_logic_vector(RNG_SEL);
    adr : std_logic_vector(RNG_ADR);
    dat : std_logic_vector(RNG_DAT);
  end record;
  constant MASTER_DEFAULT : rec_master_port :=
  (
    cyc => '0',
    stb => '0',
    we => '0',
    sel => (others => '0'),
    adr => (others => '0'),
    dat => (others => '0')
  );
  type arr_master_port is array(natural range <>) of rec_master_port;

  -- ports driven by WISHBONE slave modules
  type rec_slave_port is record
    ack : std_logic;
    dat : std_logic_vector(RNG_DAT);
  end record;
  constant SLAVE_DEFAULT : rec_slave_port :=
  (
    ack => '0',
    dat => (others => '0')
  );
  type arr_slave_port is array(natural range <>) of rec_slave_port;

  -- the whole WISHBONE signal
--  type rec_wishbone_signal is record
--    syscon : rec_syscon_port;
--    master : rec_master_port;
--    slave : rec_slave_port;
--  end record;
--  constant WISHBONE_DEFAULT : rec_wishbone_signal :=
--  (
--    syscon => SYSCON_DEFAULT,
--    master => MASTER_DEFAULT,
--    slave => SLAVE_DEFAULT
--  );

  -- signals connecting masters to intercon module
  type rec_masters_signal is record
    --syscon : arr_syscon_port(RNG_MASTERS);
    master : arr_master_port(RNG_MASTERS);
    slave : arr_slave_port(RNG_MASTERS);
  end record;
  constant MASTERS_DEFAULT : rec_masters_signal :=
  (
    --syscon => (others => SYSCON_DEFAULT),
    master => (others => MASTER_DEFAULT),
    slave => (others => SLAVE_DEFAULT)
  );

  -- signals connecting slaves to intercon module
  type rec_slaves_signal is record
    --syscon : arr_syscon_port(RNG_SLAVES);
    master : arr_master_port(RNG_SLAVES);
    slave : arr_slave_port(RNG_SLAVES);
  end record;
  constant SLAVES_DEFAULT : rec_slaves_signal :=
  (
    --syscon => (others => SYSCON_DEFAULT),
    master => (others => MASTER_DEFAULT),
    slave => (others => SLAVE_DEFAULT)
  );

  -- all signals connected to intercon module
  type rec_intercon_signal is record
    --syscon : rec_syscon_port;
    masters : rec_masters_signal;
    slaves : rec_slaves_signal;
  end record;
  constant INTERCON_DEFAULT : rec_intercon_signal :=
  (
    --syscon => SYSCON_DEFAULT,
    masters => MASTERS_DEFAULT,
    slaves => SLAVES_DEFAULT
  );
  
  -- WISHBONE modules
  -------------------

  -- interconnection logic
  component wb_intercon
    generic (
      NR_OF_DBGPORTS : positive);
    port (
      syscon_i : in  rec_syscon_port;
      master_i : in  arr_master_port(RNG_MASTERS);
      slave_o  : out arr_slave_port(RNG_MASTERS);
      master_o : out arr_master_port(RNG_SLAVES);
      slave_i  : in  arr_slave_port(RNG_SLAVES);
      debug    : out std_logic_vector((NR_OF_DBGPORTS-1) downto 0));
  end component;
  
  -- a single signal connects all WISHBONE devices to intercon
  signal intercon : rec_intercon_signal := INTERCON_DEFAULT;

  -- partial address decoding is done in the WISHBONE system
  -- some msbs of the address are used to select the appropriate slave device
  -- in the interconnection logic (base address decoding)
  -- an unique slave id is derived from the unsigned integer interpretation of these bits
  -- the slave id is used as an index and identifies the slave module in array data types
  -- the master id is assigned manually
  -- some lsbs of the address are used to select which memory unit or parameter to read
  -- or write inside slave devices (address offset decoding)
  -- FPGA resources can be saved by decoding only the necessary bits
  -- example: if you have five parameters in your slave device,
  -- then you decode only the three lsbs
  -- absolute address := base address + offset
  subtype RNG_SLAVE_SELECT is natural range 30 downto 28;

  -- bfm subroutines (for simulation purposes only!)
  --------------------------------------------------
  
  -- synthesis translate_off
      
  type arr_std32 is array(natural range <>) of std_logic_vector(31 downto 0);
  
  signal bfm_ma : rec_master_port := MASTER_DEFAULT;
  signal bfm_sl : rec_slave_port := SLAVE_DEFAULT;
  signal bfm_sys : rec_syscon_port := SYSCON_DEFAULT;

  procedure wb_write (
    addr : in std_logic_vector(31 downto 0); -- initial wb address
    data : in arr_std32; -- data buffer to be transfered
    no_cycles : in positive; -- number of 32-bit word transactions, not byte-count!
    addr_incr : in boolean; -- auto increment address in block transfers
    signal ma2wb : out rec_master_port;
    signal wb2ma : in rec_slave_port;
    signal sys : in rec_syscon_port
	);
  
  procedure wb_read (
    addr : in std_logic_vector(31 downto 0); -- initial wb address
    data : out arr_std32; -- data buffer to be transfered
    no_cycles : in positive; -- number of 32-bit word transactions, not byte-count!
    addr_incr : in boolean; -- auto increment address in block transfers
    signal ma2wb : out rec_master_port;
    signal wb2ma : in rec_slave_port;
    signal sys : in rec_syscon_port
	);
  
  procedure wb_write_reg (
    addr : in std_logic_vector(31 downto 0); -- initial wb address
    data : in std_logic_vector(RNG_DAT); -- data to be transfered
    signal ma2wb : out rec_master_port;
    signal wb2ma : in rec_slave_port;
    signal sys : in rec_syscon_port
	);
  
  procedure wb_read_reg (
    addr : in std_logic_vector(31 downto 0); -- initial wb address
    data : out std_logic_vector(RNG_DAT); -- data to be transfered
    signal ma2wb : out rec_master_port;
    signal wb2ma : in rec_slave_port;
    signal sys : in rec_syscon_port
	);
  
  procedure wb_read_reg (
    addr : in std_logic_vector(31 downto 0); -- initial wb address
    signal ma2wb : out rec_master_port;
    signal wb2ma : in rec_slave_port;
    signal sys : in rec_syscon_port
	);
      
  -- synthesis translate_on
	  
end wishbone_pkg;


package body wishbone_pkg is

  -- synthesis translate_off

  procedure wb_write (
    addr : in std_logic_vector(31 downto 0); -- initial wb address
    data : in arr_std32; -- data buffer to be transfered
    no_cycles : in positive; -- number of 32-bit word transactions, not byte-count!
    addr_incr : in boolean; -- auto increment address in block transfers
    signal ma2wb : out rec_master_port;
    signal wb2ma : in rec_slave_port;
    signal sys : in rec_syscon_port
	) is
  begin
    ma2wb.sel <= (others => '1');
    for i in 0 to (no_cycles-1) loop
      wait until sys.clk = '1';
      wait for 1 ns;
      ma2wb.cyc <= '1';
      ma2wb.stb <= '1';
      ma2wb.we <= '1';
      if addr_incr then
        ma2wb.adr <= std_logic_vector(unsigned(addr(RNG_ADR))+i);
      else
        ma2wb.adr <= addr(RNG_ADR);
      end if;
      ma2wb.dat <= data(i);
      while true loop
        wait until sys.clk = '0';
        if wb2ma.ack = '1' then
          exit;
        end if;
      end loop;
    end loop;
    wait until sys.clk = '1';
    wait for 1 ns;
    ma2wb.cyc <= '0';
    ma2wb.stb <= '0';
  end wb_write;
  
  procedure wb_read (
    addr : in std_logic_vector(31 downto 0); -- initial wb address
    data : out arr_std32; -- data buffer to be transfered
    no_cycles : in positive; -- number of 32-bit word transactions, not byte-count!
    addr_incr : in boolean; -- auto increment address in block transfers
    signal ma2wb : out rec_master_port;
    signal wb2ma : in rec_slave_port;
    signal sys : in rec_syscon_port
	) is
  begin
    ma2wb.sel <= (others => '1');
    for i in 0 to (no_cycles-1) loop
      wait until sys.clk = '1';
      wait for 1 ns;
      ma2wb.cyc <= '1';
      ma2wb.stb <= '1';
      ma2wb.we <= '0';
      if addr_incr then
        ma2wb.adr <= std_logic_vector(unsigned(addr(RNG_ADR))+i);
      else
        ma2wb.adr <= addr(RNG_ADR);
      end if;
      while true loop
        wait until sys.clk = '0';
        if wb2ma.ack = '1' then
          exit;
        end if;
      end loop;
      data(i) := wb2ma.dat;
    end loop;
    wait until sys.clk = '1';
    wait for 1 ns;
    ma2wb.cyc <= '0';
    ma2wb.stb <= '0';
  end wb_read;
  
  procedure wb_write_reg (
    addr : in std_logic_vector(31 downto 0); -- wb address
    data : in std_logic_vector(RNG_DAT); -- data to be transfered
    signal ma2wb : out rec_master_port;
    signal wb2ma : in rec_slave_port;
    signal sys : in rec_syscon_port
	) is
    variable wr_reg_data : arr_std32(0 downto 0) := (others => (others => '-'));
  begin
    wr_reg_data(0) := data;
    wb_write(addr, wr_reg_data, 1, false, ma2wb, wb2ma, sys);
  end wb_write_reg;
  
  procedure wb_read_reg (
    addr : in std_logic_vector(31 downto 0); -- wb address
    data : out std_logic_vector(RNG_DAT); -- data to be transfered
    signal ma2wb : out rec_master_port;
    signal wb2ma : in rec_slave_port;
    signal sys : in rec_syscon_port
	) is
    variable rd_reg_data : arr_std32(0 downto 0) := (others => (others => '-'));
  begin
    wb_read(addr, rd_reg_data, 1, false, ma2wb, wb2ma, sys);
    data := rd_reg_data(0);
  end wb_read_reg;
  
  procedure wb_read_reg (
    addr : in std_logic_vector(31 downto 0); -- wb address
    signal ma2wb : out rec_master_port;
    signal wb2ma : in rec_slave_port;
    signal sys : in rec_syscon_port
	) is
    variable rd_reg_data : arr_std32(0 downto 0) := (others => (others => '-'));
  begin
    wb_read(addr, rd_reg_data, 1, false, ma2wb, wb2ma, sys);
  end wb_read_reg;
  
  -- synthesis translate_on
  
end wishbone_pkg;
